@@ -5,6 +5,8 @@ BEGIN { push(@INC, ".."); };
 use WebminCore;
 $directive_type_count = 20;
 
+our ($saved_conf_files);
+
 if ($module_name ne 'htaccess') {
 	&init_config();
 	%access = &get_module_acl();
@@ -16,76 +18,102 @@ else {
 	}
 map { $access_types{$_}++ } @access_types;
 $site_file = ($config{'webmin_apache'} || $module_config_directory)."/site";
+$httpd_info_cache = $module_config_directory."/httpd-info";
+$last_config_change_flag = $module_var_directory."/config-flag";
+$last_restart_time_flag = $module_var_directory."/restart-flag";
 
 # Check if a list of supported modules needs to be built. This is done
 # if the Apache binary changes, when Webmin is upgraded, or once every five
 # minutes if automatic rebuilding is enabled.
 if ($module_name ne 'htaccess') {
-	local %oldsite;
-	local $httpd = &find_httpd();
-	local @st = stat($httpd);
-	&read_file($site_file, \%oldsite);
-	local @sst = stat($site_file);
-	if ($oldsite{'path'} ne $httpd ||
-	    $oldsite{'size'} != $st[7] ||
-	    $oldsite{'webmin'} != &get_webmin_version() ||
-	    $config{'auto_mods'} && $sst[9] < time()-5*60) {
-		# Need to build list of supported modules
-		local ($ver, $mods) = &httpd_info($httpd);
-		if ($ver) {
-			local @mods = map { "$_/$ver" } &configurable_modules();
-			foreach my $m (@mods) {
-				if ($m =~ /(\S+)\/(\S+)/) {
-					$httpd_modules{$1} = $2;
-					}
-				}
-			# Call again now that known modules have been set, as
-			# sometimes there are dependencies due to LoadModule
-			# statements in an IfModule block
-			@mods = map { "$_/$ver" } &configurable_modules();
-			local %site = ( 'size' => $st[7],
-					'path' => $httpd,
-					'modules' => join(' ', @mods),
-					'webmin' => &get_webmin_version() );
-			&lock_file($site_file);
-			&write_file($site_file, \%site);
-			chmod(0644, $site_file);
-			&unlock_file($site_file);
-			}
-		}
-	}
-
-# Read the site-specific setup file, then require in all the module-specific
-# .pl files
-if (&read_file($site_file, \%site)) {
-	local($m, $f, $d);
-	$httpd_size = $site{'size'};
-	foreach $m (split(/\s+/, $site{'modules'})) {
-		if ($m =~ /(\S+)\/(\S+)/) {
-			$httpd_modules{$1} = $2;
-			}
-		}
-	foreach $m (keys %httpd_modules) {
-		if (!-r "$module_root_directory/$m.pl") {
-			delete($httpd_modules{$m});
-			}
-		}
-	foreach $f (split(/\s+/, $site{'htaccess'})) {
-		if (-r $f) { push(@htaccess_files, $f); }
-		}
-	foreach $m (keys %httpd_modules) {
-		do "$m.pl";
-		}
-	foreach $d (split(/\s+/, $site{'defines'})) {
-		$httpd_defines{$d}++;
-		}
+	&create_site_file();
 	}
+&read_site_file();
 
 $apache_docbase = $config{'apache_docbase'} ? $config{'apache_docbase'} :
 		  $httpd_modules{'core'} >= 2.0 ?
 			"http://httpd.apache.org/docs-2.0/mod/" :
 			"http://httpd.apache.org/docs/mod/";
 
+# create_site_file()
+# If the Apache binary or Webmin version has changed, create the site
+# file containing all known Apache modules
+sub create_site_file
+{
+my %oldsite;
+my $httpd = &find_httpd();
+my @st = stat($httpd);
+&read_file($site_file, \%oldsite);
+my @sst = stat($site_file);
+if ($oldsite{'path'} ne $httpd ||
+    $oldsite{'size'} != $st[7] ||
+    $oldsite{'webmin'} != &get_webmin_version() ||
+    $config{'auto_mods'} && $sst[9] < time()-5*60) {
+	# Need to build list of supported modules
+	my ($ver, $mods, $fullver) = &httpd_info($httpd);
+	if ($ver) {
+		my @allmods = &available_modules();
+		my @mods = map { "$_/$ver" }
+			       &configurable_modules(\@allmods);
+		foreach my $m (@mods) {
+			if ($m =~ /(\S+)\/(\S+)/) {
+				$httpd_modules{$1} = $2;
+				}
+			}
+		# Call again now that known modules have been set, as
+		# sometimes there are dependencies due to LoadModule
+		# statements in an IfModule block
+		undef(@get_config_cache);
+		@allmods = &available_modules();
+		@mods = map { "$_/$ver" }
+			    &configurable_modules(\@allmods);
+		my %site = ( 'size' => $st[7],
+			     'path' => $httpd,
+			     'modules' => join(' ', @mods),
+			     'allmodules' => join(' ', @allmods),
+			     'version' => $ver,
+			     'fullversion' => $fullver,
+			     'webmin' => &get_webmin_version() );
+		&lock_file($site_file);
+		&write_file($site_file, \%site);
+		chmod(0644, $site_file);
+		&unlock_file($site_file);
+		}
+	}
+}
+
+# read_site_file()
+# Read the site-specific setup file, then require in all the module-specific
+# .pl files
+sub read_site_file
+{
+if (&read_file($site_file, \%site)) {
+	foreach my $m (split(/\s+/, $site{'modules'})) {
+		if ($m =~ /(\S+)\/(\S+)/) {
+			$httpd_modules{$1} = $2;
+			}
+		}
+	foreach my $m (split(/\s+/, $site{'allmodules'})) {
+		$all_httpd_modules{$m} = $site{'version'};
+		}
+	foreach my $m (keys %httpd_modules) {
+		if (!-r "$module_root_directory/$m.pl") {
+			delete($httpd_modules{$m});
+			}
+		}
+	foreach my $f (split(/\s+/, $site{'htaccess'})) {
+		if (-r $f) { push(@htaccess_files, $f); }
+		}
+	foreach my $m (keys %httpd_modules) {
+		do "$m.pl";
+		}
+	foreach my $d (split(/\s+/, $site{'defines'})) {
+		$httpd_defines{$d}++;
+		}
+	}
+
+}
+
 # parse_config_file(handle, lines, file, [recursive])
 # Parses lines of text from some config file into a data structure. The
 # return value is an array of references, one for each directive in the file.
@@ -97,6 +125,7 @@ $apache_docbase = $config{'apache_docbase'} ? $config{'apache_docbase'} :
 #  name -	The name of this directive
 #  value -	Value (possibly with spaces)
 #  members -	For type 1, a reference to the array of members
+#  indent -     Number of spaces before the name
 sub parse_config_file
 {
 local($fh, @rv, $line, %dummy);
@@ -116,7 +145,7 @@ foreach my $d (&get_httpd_defines()) {
 		}
 	}
 while($line = <$fh>) {
-	chop;
+	$line =~ s/\r|\n//g;
 	$line =~ s/^\s*#.*$//g;
 	if ($line =~ /^\s*<\/(\S+)\s*(.*)>/) {
 		# end of a container directive. This can only happen in a
@@ -136,6 +165,11 @@ while($line = <$fh>) {
 		$altmod =~ s/^(\S+)_module$/mod_$1/g;
 		local $mpmmod = $mod;
 		$mpmmod =~ s/^mpm_//; $mpmmod =~ s/_module$//;
+		if ($mod eq "prefork") {
+			# Special case for all the prefork aliases
+			($mod, $altmod, $mpmmod) = ("mod_".$mod."_module",
+						    "mod_mpm_".$mod, $mod);
+			}
 		if (!$not && $httpd_modules{$mod} ||
 		    $not && !$httpd_modules{$mod} ||
 		    !$not && $httpd_modules{$altmod} ||
@@ -224,32 +258,38 @@ while($line = <$fh>) {
 				    'name', "</IfVersion>" });
 			}
 		}
-	elsif ($line =~ /^\s*<(\S+)\s*(.*)>/) {
+	elsif ($line =~ /^(\s*)<(\S+)\s*(.*)>/) {
 		# start of a container directive. The first member is a dummy
 		# directive at the same line as the container
 		local(%dir, @members);
 		%dir = ('line', $_[1],
 			'file', $_[2],
 			'type', 1,
-			'name', $1,
-			'value', $2);
+			'name', $2,
+			'value', $3);
+		local $indent = $1;
 		$dir{'value'} =~ s/\s+$//g;
 		$dir{'words'} = &wsplit($dir{'value'});
 		$_[1]++;
 		@members = &parse_config_file($fh, $_[1], $_[2], $dir{'name'});
 		$dir{'members'} = \@members;
 		$dir{'eline'} = $_[1]-1;
+		$indent =~ s/\t/        /g;
+		$dir{'indent'} = length($indent);
 		push(@rv, \%dir);
 		}
-	elsif ($line =~ /^\s*(\S+)\s*(.*)$/) {
+	elsif ($line =~ /^(\s*)(\S+)\s*(.*)$/) {
 		# normal directive
 		local(%dir);
 		%dir = ('line', $_[1],
 			'eline', $_[1],
 			'file', $_[2],
 			'type', 0,
-			'name', $1,
-			'value', $2);
+			'name', $2,
+			'value', $3);
+		local $indent = $1;
+		$indent =~ s/\t/        /g;
+		$dir{'indent'} = length($indent);
 		if ($dir{'value'} =~ s/\\$//g) {
 			# multi-line directive!
 			while($line = <$fh>) {
@@ -318,7 +358,7 @@ return wantarray ? @vals : !@vals ? undef : $vals[$#vals];
 }
 
 # find_directive_struct(name, &directives)
-# Returns references to directives maching some name
+# Returns references to directives matching some name
 sub find_directive_struct
 {
 local (@vals, $ref);
@@ -381,7 +421,7 @@ push(@get_config_cache, &get_config_file($acc, \%seenfiles));
 foreach $v (@virt) {
 	my %seenfiles;
 	$mref = $v->{'members'};
-	foreach $idn ("ResourceConfig", "AccessConfig", "Include") {
+	foreach $idn ("ResourceConfig", "AccessConfig", "Include", "IncludeOptional") {
 		foreach $inc (&find_directive_struct($idn, $mref)) {
 			local @incs = &expand_apache_include(
 					$inc->{'words'}->[0]);
@@ -424,7 +464,8 @@ else {
 	}
 
 # Expand Include directives
-foreach $inc (&find_directive_struct("Include", \@rv)) {
+foreach $inc (&find_directive_struct("Include", \@rv),
+	      &find_directive_struct("IncludeOptional", \@rv)) {
 	local @incs = &expand_apache_include($inc->{'words'}->[0]);
 	foreach my $ginc (@incs) {
 		push(@rv, &get_config_file($ginc, $seen));
@@ -520,7 +561,7 @@ local($i, @old, $lref, $change, $len, $v);
 @old = &find_directive_struct($_[0], $_[2]);
 local @files;
 for($i=0; $i<@old || $i<@{$_[1]}; $i++) {
-	$v = ${$_[1]}[$i];
+	$v = $i<@{$_[1]} ? $_[1]->[$i] : undef;
 	if ($i >= @old) {
 		# a new directive is being added. If other directives of this
 		# type exist, add it after them. Otherwise, put it at the end of
@@ -532,6 +573,7 @@ for($i=0; $i<@old || $i<@{$_[1]}; $i++) {
 			%v = (	"line", $change->{'line'}+1,
 				"eline", $change->{'line'}+1,
 				"file", $change->{'file'},
+				"indent", $change->{'indent'},
 				"type", 0,
 				"name", $_[0],
 				"value", $v,
@@ -541,7 +583,8 @@ for($i=0; $i<@old || $i<@{$_[1]}; $i++) {
 			splice(@{$_[2]}, $j, 0, \%v);
 			$lref = &read_file_lines($v{'file'});
 			push(@files, $v{'file'});
-			splice(@$lref, $v{'line'}, 0, "$_[0] $v");
+			splice(@$lref, $v{'line'}, 0,
+				(" " x $v{'indent'})."$_[0] $v");
 			$change = \%v;
 			}
 		else {
@@ -549,20 +592,30 @@ for($i=0; $i<@old || $i<@{$_[1]}; $i++) {
 			# in this section
 			local($f, %v, $j);
 			$f = $_[2]->[0]->{'file'};
-			for($j=0; $_[2]->[$j]->{'file'} eq $f; $j++) { }
-			$l = $_[2]->[$j-1]->{'eline'}+1;
+			for($j=0; $j < @{$_[2]} &&
+				  $_[2]->[$j]->{'file'} eq $f; $j++) { }
+			$lref = &read_file_lines($f);
+			if ($_[2] eq $_[3]) {
+				# Top-level, so add to the end of the file
+				$l = scalar(@$lref) + 1;
+				}
+			else {
+				# Add after last directive in the same section
+				$l = $_[2]->[$j-1]->{'eline'}+1;
+				}
 			%v = (	"line", $l,
 				"eline", $l,
 				"file", $f,
+				"indent", $_[2]->[$j-1]->{'indent'},
 				"type", 0,
 				"name", $_[0],
 				"value", $v,
 				"words", &wsplit($v) );
 			&renumber($_[3], $l, $f, 1);
 			splice(@{$_[2]}, $j, 0, \%v);
-			$lref = &read_file_lines($f);
 			push(@files, $f);
-			splice(@$lref, $l, 0, "$_[0] $v");
+			splice(@$lref, $l, 0, 
+				(" " x $v{'indent'})."$_[0] $v");
 			}
 		}
 	elsif ($i >= @{$_[1]}) {
@@ -585,11 +638,15 @@ for($i=0; $i<@old || $i<@{$_[1]}; $i++) {
 		$old[$i]->{'value'} = $v;
 		$old[$i]->{'words'} = &wsplit($v);
 		$old[$i]->{'eline'} = $old[$i]->{'line'};
-		splice(@$lref, $old[$i]->{'line'}, $len, "$_[0] $v");
+		splice(@$lref, $old[$i]->{'line'}, $len,
+			(" " x $old[$i]->{'indent'}).$_[0]." ".$v);
 		$change = $old[$i];
 		}
 	}
-return &unique(@files);
+&update_last_config_change();
+@files = &unique(@files);
+push(@{$saved_conf_files}, @files);
+return @files;
 }
 
 # save_directive_struct(&old-directive, &directive, &parent-directives,
@@ -603,15 +660,26 @@ local $file = $olddir ? $olddir->{'file'} :
 	      $newdir->{'file'} ? $newdir->{'file'} : $pconf->[0]->{'file'};
 local $lref = &read_file_lines($file);
 local $oldlen = $olddir ? $olddir->{'eline'}-$olddir->{'line'}+1 : undef;
-local @newlines = $newdir ? &directive_lines($newdir) : ( );
+local @newlines;
+if ($newdir) {
+	my $isrc = $olddir ? $olddir :
+		   @$pconf ? $pconf->[0] : undef;
+	if ($isrc) {
+		&recursive_set_indent($newdir, $isrc->{'indent'});
+		}
+	$newdir{'words'} = &wsplit($newdir{'value'});
+	@newlines = &directive_lines($newdir);
+	}
 if ($olddir && $newdir) {
 	# Update in place
+	$newdir->{'words'} = &wsplit($newdir->{'value'});
 	if ($first) {
 		# Just changing first and last line, like virtualhost IP
 		$lref->[$olddir->{'line'}] = $newlines[0];
 		$lref->[$olddir->{'eline'}] = $newlines[$#newlines];
 		$olddir->{'name'} = $newdir->{'name'};
 		$olddir->{'value'} = $newdir->{'value'};
+		$olddir->{'words'} = $newdir->{'words'};
 		}
 	else {
 		# Re-writing whole block
@@ -655,6 +723,7 @@ elsif (!$olddir && $newdir) {
 		$addpos--;
 		$addline = $pconf->[$addpos]->{'eline'}+1;
 		}
+	$newdir->{'words'} = &wsplit($newdir->{'value'});
 	$newdir->{'file'} = $file;
 	$newdir->{'line'} = $addline;
 	$newdir->{'eline'} = $addline + scalar(@newlines) - 1;
@@ -669,26 +738,38 @@ elsif (!$olddir && $newdir) {
 					   $newdir->{'file'});
 		}
 	}
+&update_last_config_change();
+}
+
+# recursive_set_indent(&directive, indent)
+# Set the indent to match another at the same level
+sub recursive_set_indent
+{
+my ($dir, $indent) = @_;
+$dir->{'indent'} = $indent if (!defined($dir->{'indent'}));
+if ($dir->{'type'}) {
+	foreach my $m (@{$dir->{'members'}}) {
+		&recursive_set_indent($m, $indent+4);
+		}
+	}
 }
 
 # recursive_set_lines_files(&directives, first-line, file)
 # Update the line numbers and filenames in a list of directives
 sub recursive_set_lines_files
 {
-local ($dirs, $line, $file) = @_;
-foreach my $d (@$dirs) {
+my ($dirs, $line, $file) = @_;
+foreach my $dir (@$dirs) {
 	$dir->{'line'} = $line;
 	$dir->{'file'} = $file;
 	if ($dir->{'type'}) {
 		# Do sub-members too
 		&recursive_set_lines_files($dir->{'members'}, $line+1, $file);
-		$line += scalar(@{$dir->{'members'}})+1;
-		$dir->{'eline'} = $line;
+		$line += scalar(grep { $_->{'name'} ne 'dummy' }
+				     @{$dir->{'members'}})+1;
 		}
-	else {
-		$dir->{'eline'} = $line;
-		}
-	$line++;
+	$dir->{'eline'} = $line;
+	$line++ if ($dir->{'name'} ne 'dummy');
 	}
 return $line;
 }
@@ -712,22 +793,23 @@ unlink($file);
 # beyond the given line.
 sub renumber
 {
-local($d);
-if (!$_[3]) { return; }
-foreach $d (@{$_[0]}) {
-	if ($d->{'file'} eq $_[2] && $d->{'line'} >= $_[1]) {
-		$d->{'line'} += $_[3];
+my ($conf, $line, $file, $offset) = @_;
+return if (!$offset);
+foreach my $d (@$conf) {
+	if ($d->{'file'} eq $file && $d->{'line'} >= $_[1]) {
+		$d->{'line'} += $offset;
 		}
-	if ($d->{'file'} eq $_[2] && $d->{'eline'} >= $_[1]) {
-		$d->{'eline'} += $_[3];
+	if ($d->{'file'} eq $file && $d->{'eline'} >= $_[1]) {
+		$d->{'eline'} += $offset;
 		}
 	if ($d->{'type'}) {
-		&renumber($d->{'members'}, $_[1], $_[2], $_[3]);
+		&renumber($d->{'members'}, $line, $file, $offset);
 		}
 	}
 }
 
-# server_root(path, &directives)
+# server_root(path)
+# Convert a relative path to being under the server root
 sub server_root
 {
 if (!$_[0]) { return undef; }
@@ -759,21 +841,21 @@ return $_[0] ? $_[0] : $_[1];
 }
 
 # make_directives(ref, version, module)
+# Return directives suitable for this system and version
 sub make_directives
 {
-local(@rv, $aref);
-$aref = $_[0];
-local $ver = $_[1];
+my ($aref, $ver, $mod) = @_;
+my @rv;
 if ($ver =~ /^(1)\.(3)(\d+)$/) {
 	$ver = sprintf "%d.%d%2.2d", $1, $2, $3;
 	}
-foreach $d (@$aref) {
+foreach my $d (@$aref) {
 	local(%dir);
 	$dir{'name'} = $d->[0];
 	$dir{'multiple'} = $d->[1];
 	$dir{'type'} = int($d->[2]);
 	$dir{'subtype'} = $d->[2] - $dir{'type'};
-	$dir{'module'} = $_[2];
+	$dir{'module'} = $mod;
 	$dir{'version'} = $ver;
 	$dir{'priority'} = $d->[5];
 	foreach $c (split(/\s+/, $d->[3])) { $dir{$c}++; }
@@ -795,7 +877,9 @@ sub editable_directives
 local($m, $func, @rv, %done);
 foreach $m (keys %httpd_modules) {
 	$func = $m."_directives";
-	push(@rv, &$func($httpd_modules{$m}));
+	if (defined(&$func)) {
+		push(@rv, &$func($httpd_modules{$m}));
+		}
 	}
 @rv = grep { $_->{'type'} == $_[0] && $_->{$_[1]} &&
 	     !$done{$_->{'name'}}++ } @rv;
@@ -859,8 +943,8 @@ foreach $e (@{$_[0]}) {
 		$names = " (";
 		foreach $ed (split(/\s+/, $e->{'name'})) {
 			# nodo50 v0.1 - Change 000004 - Open new window for Help in Apache module and mod_apachessl Help from http://www.apache-ssl.org and
-			# nodo50 v0.1 - Change 000004 - Abre nueva ventana para Ayuda del mdulo Apache y para mod_apachessl busca la Ayuda en http://www.apache-ssl.org and
-			$names .= "<tt><a href='".($e->{'module'} eq 'mod_apachessl' ? 'http://www.apache-ssl.org/docs.html#'.$ed : $apache_docbase."/".$e->{'module'}.".html#".lc($ed))."'>".$ed."</a></tt> ";
+			# nodo50 v0.1 - Change 000004 - Abre nueva ventana para Ayuda del módulo Apache y para mod_apachessl busca la Ayuda en http://www.apache-ssl.org and
+			$names .= "<tt>".&ui_link( ($e->{'module'} eq 'mod_apachessl' ? 'http://www.apache-ssl.org/docs.html#'.$ed : $apache_docbase."/".$e->{'module'}.".html#".lc($ed)), $ed )."</tt>&nbsp;";
 			#$names .= "<tt><a href='".$apache_docbase."/".$e->{'module'}.".html#".lc($ed)."'>".$ed."</a></tt> ";
 			# nodo50 v0.1 - Change 000004 - End
 			}
@@ -909,6 +993,7 @@ for($i=0; $i<@chname; $i++) {
 &flush_file_lines();
 &unlock_apache_files();
 &after_changing();
+&format_modifed_config_files();
 }
 
 # opt_input(value, name, default, size)
@@ -922,6 +1007,7 @@ sub parse_opt
 {
 local($i, $re);
 local $v = $in{$_[0]};
+$v =~ /\r|\n|\0/ && &error($text{'enewline'});
 if ($in{"$_[0]_def"}) { return ( [ ] ); }
 for($i=1; $i<@_; $i+=2) {
 	$re = $_[$i];
@@ -934,7 +1020,7 @@ return ( [ $v =~ /\s/ && !$_[3] ? "\"$v\"" : $v ] );
 # Each choice is a display,value pair
 sub choice_input
 {
-local($i, $rv);
+my($i, $rv);
 for($i=3; $i<@_; $i++) {
 	$_[$i] =~ /^([^,]*),(.*)$/;
 	$rv .= &ui_oneradio($_[1], $2, $1, lc($2) eq lc($_[0]) ||
@@ -947,12 +1033,11 @@ return $rv;
 # Each choice is a display,value pair
 sub choice_input_vert
 {
-local($i, $rv);
+my($i, $rv);
 for($i=3; $i<@_; $i++) {
 	$_[$i] =~ /^([^,]*),(.*)$/;
-	$rv .= sprintf "<input type=radio name=$_[1] value=\"$2\" %s> $1<br>\n",
-		lc($2) eq lc($_[0]) || !defined($_[0]) &&
-				       lc($2) eq lc($_[2]) ? "checked" : "";
+	$rv .= &ui_oneradio($_[1], $2, $1, lc($2) eq lc($_[0]) ||
+				!defined($_[0]) && lc($2) eq lc($_[2]))."<br>\n";
 	}
 return $rv;
 }
@@ -960,22 +1045,28 @@ return $rv;
 # parse_choice(name, default)
 sub parse_choice
 {
-if (lc($in{$_[0]}) eq lc($_[1])) { return ( [ ] ); }
-else { return ( [ $in{$_[0]} ] ); }
+if (lc($in{$_[0]}) eq lc($_[1])) {
+	return ( [ ] );
+	}
+else {
+	$in{$_[0]} =~ /\r|\n|\0/ && &error($text{'enewline'});
+	return ( [ $in{$_[0]} ] );
+	}
 }
 
 # select_input(value, name, default, [choice]+)
 sub select_input
 {
-local($i, $rv);
-$rv = "<select name=\"$_[1]\">\n";
+my($i, @sel);
+my $selv;
 for($i=3; $i<@_; $i++) {
 	$_[$i] =~ /^([^,]*),(.*)$/;
-	$rv .= sprintf "<option value=\"$2\" %s> $1\n",
-		lc($2) eq lc($_[0]) || !defined($_[0]) && lc($2) eq lc($_[2]) ? "selected" : "";
+	if (lc($2) eq lc($_[0]) || !defined($_[0]) && lc($2) eq lc($_[2])) {
+		$selv = $2;
+		}
+	push(@sel, [ $2, $1 || "&nbsp;" ]);
 	}
-$rv .= "</select>\n";
-return $rv;
+return &ui_select($_[1], $selv, \@sel, 1);
 }
 
 # parse_choice(name, default)
@@ -987,8 +1078,8 @@ return &parse_choice(@_);
 # handler_input(value, name)
 sub handler_input
 {
-local($m, $func, @hl, $rv, $h);
-local $conf = &get_config();
+my($m, $func, @hl, @sel, $h);
+my $conf = &get_config();
 push(@hl, "");
 foreach $m (keys %httpd_modules) {
 	$func = $m."_handlers";
@@ -997,22 +1088,23 @@ foreach $m (keys %httpd_modules) {
 		}
 	}
 if (&indexof($_[0], @hl) < 0) { push(@hl, $_[0]); }
-$rv = "<select name=$_[1]>\n";
 foreach $h (&unique(@hl)) {
-	$rv .= sprintf "<option value=\"$h\" %s>$h\n",
-		$h eq $_[0] ? "selected" : "";
+    push(@sel, [$h, $h, ($h eq $_[0] ? "selected" : "")] );
 	}
-$rv .= sprintf "<option value=\"None\" %s>&lt;$text{'core_none'}&gt;\n",
-	$_[0] eq "None" ? "selected" : "";
-$rv .= "</select>\n";
-return $rv;
+push(@sel, ["None", "&lt;".$text{'core_none'}."&gt;", ($_[0] eq "None" ? "selected" : "")] );
+return &ui_select($_[1], undef, \@sel, 1);
 }
 
 # parse_handler(name)
 sub parse_handler
 {
-if ($in{$_[0]} eq "") { return ( [ ] ); }
-else { return ( [ $in{$_[0]} ] ); }
+if ($in{$_[0]} eq "") {
+	return ( [ ] );
+	}
+else {
+	$in{$_[0]} =~ /\r|\n|\0/ && &error($text{'enewline'});
+	return ( [ $in{$_[0]} ] );
+	}
 }
 
 # filters_input(&values, name)
@@ -1030,15 +1122,16 @@ foreach $f (@{$_[0]}) {
 	push(@fl, $f) if (&indexof($f, @fl) < 0);
 	}
 foreach $f (&unique(@fl)) {
-	$rv .= sprintf "<input type=checkbox name=$_[1] value='%s' %s> %s\n",
-			$f, &indexof($f, @{$_[0]}) < 0 ? "" : "checked", $f;
+	$rv .= &ui_checkbox($_[1], $f, $f, (&indexof($f, @{$_[0]}) < 0 ? 0 : 1 ) ); 
 	}
+$rv ||= &ui_textbox($_[1], "", 20);
 return $rv;
 }
 
 # parse_filters(name)
 sub parse_filters
 {
+$in{$_[0]} =~ /\r|\n|\0/ && &error($text{'enewline'});
 local @f = split(/\0/, $in{$_[0]});
 return @f ? ( [ join(";", @f) ] ) : ( [ ] );
 }
@@ -1106,17 +1199,20 @@ close(USERS);
 # some context (global, virtual, directory or htaccess)
 sub config_icons
 {
+local ($ctx, $prog) = @_;
 local($m, $func, $e, %etype, $i, $c);
 foreach $m (sort { $a cmp $b } (keys %httpd_modules)) {
         $func = $m."_directives";
-	foreach $e (&$func($httpd_modules{$m})) {
-		if ($e->{$_[0]}) { $etype{$e->{'type'}}++; }
+	if (defined(&$func)) {
+		foreach $e (&$func($httpd_modules{$m})) {
+			if ($e->{$ctx}) { $etype{$e->{'type'}}++; }
+			}
 		}
         }
 local (@titles, @links, @icons);
 for($i=0; $text{"type_$i"}; $i++) {
 	if ($etype{$i} && $access_types{$i}) {
-		push(@links, $_[1]."type=$i");
+		push(@links, $prog."type=$i");
 		push(@titles, $text{"type_$i"});
 		push(@icons, "images/type_icon_$i.gif");
 		}
@@ -1136,19 +1232,25 @@ print "<p>\n";
 # Returns HTML for a link to put in the top-right corner of every page
 sub restart_button
 {
-local $rv;
-$args = "redir=".&urlize(&this_url());
+local $args = "redir=".&urlize(&this_url());
 local @rv;
 if (&is_apache_running()) {
-	if ($access{'apply'}) {
-		push(@rv, "<a href=\"restart.cgi?$args\">$text{'apache_apply'}</a>\n");
-		}
 	if ($access{'stop'}) {
-		push(@rv, "<a href=\"stop.cgi?$args\">$text{'apache_stop'}</a>\n");
+		push(@rv, &ui_link("stop.cgi?$args", $text{'apache_stop'}) );
+		}
+	if ($access{'apply'}) {
+		my $n = &needs_config_restart();
+		if ($n) {
+			push(@rv, &ui_link("restart.cgi?$args&newconfig=1",
+					"<b>$text{'apache_apply'}</b>") );
+			}
+		else {
+			push(@rv, &ui_link("restart.cgi?$args", $text{'apache_apply'}) );
+			}
 		}
 	}
 elsif ($access{'stop'}) {
-	push(@rv, "<a href=\"start.cgi?$args\">$text{'apache_start'}</a>\n");
+	push(@rv, &ui_link("start.cgi?$args", $text{'apache_start'}) );
 	}
 return join("<br>\n", @rv);
 }
@@ -1181,32 +1283,56 @@ return @rv;
 # Returns the httpd version and modules array
 sub httpd_info
 {
-local(@mods, $verstr, $ver, $minor);
-$verstr = &backquote_command("\"$_[0]\" -v 2>&1");
-if ($config{'httpd_version'}) {
-	$config{'httpd_version'} =~ /(\d+)\.([\d\.]+)/;
-	$ver = $1; $minor = $2; $minor =~ s/\.//g; $ver .= ".$minor";
+local ($cmd) = @_;
+$cmd = &has_command($cmd);
+local @st = stat($cmd);
+local %cache;
+&read_file_cached($httpd_info_cache, \%cache);
+if ($cache{'cmd'} eq $cmd && $cache{'time'} == $st[9]) {
+	# Cache looks up to date
+	return ($cache{'version'}, [ split(/\s+/, $cache{'mods'}) ],
+		$cache{'fullversion'});
+	}
+local(@mods, $verstr, $ver, $minor, $fullver);
+$verstr = &backquote_command(quotemeta($cmd)." -v 2>&1");
+if ($config{'httpd_version'} =~ /(\d+)\.([\d\.]+)/) {
+	$fullver = $config{'httpd_version'};
+	$ver = $1;
+	$minor = $2;
+	$minor =~ s/\.//g;
+	$ver .= ".$minor";
 	}
 elsif ($verstr =~ /Apache(\S*)\/(\d+)\.([\d\.]+)/) {
 	# standard apache
-	$ver = $2; $minor = $3; $minor =~ s/\.//g; $ver .= ".$minor";
+	$fullver = "$2.$3";
+	$ver = $2;
+	$minor = $3;
+	$minor =~ s/\.//g;
+	$ver .= ".$minor";
 	}
 elsif ($verstr =~ /HP\s*Apache-based\s*Web\s*Server(\S*)\/(\d+)\.([\d\.]+)/) {
 	# HP's apache
-	$ver = $2; $minor = $3; $minor =~ s/\.//g; $ver .= ".$minor";
+	$fullver = "$2.$3";
+	$ver = $2;
+	$minor = $3;
+	$minor =~ s/\.//g;
+	$ver .= ".$minor";
 	}
 elsif ($verstr =~ /Red\s*Hat\s+Secure\/2\.0/i) {
 	# redhat secure server 2.0
-	$ver = 1.31;
+	$fullver = $ver = 1.31;
 	}
 elsif ($verstr =~ /Red\s*Hat\s+Secure\/3\.0/i) {
 	# redhat secure server 3.0
-	$ver = 1.39;
+	$fullver = $ver = 1.39;
 	}
 elsif (&has_command("rpm") &&
        &backquote_command("rpm -q apache 2>&1") =~ /^apache-(\d+)\.([\d\.]+)/) {
 	# got version from the RPM
-	$ver = $1; $minor = $2; $minor =~ s/\.//g; $ver .= ".$minor";
+	$fullver = $ver = $1;
+	$minor = $2;
+	$minor =~ s/\.//g;
+	$ver .= ".$minor";
 	}
 else {
 	# couldn't get version
@@ -1236,7 +1362,13 @@ else {
 		}
 	@mods = &unique(@mods);
 	}
-return ($ver, \@mods);
+$cache{'cmd'} = $cmd;
+$cache{'time'} = $st[9];
+$cache{'version'} = $ver;
+$cache{'fullversion'} = $fullver;
+$cache{'mods'} = join(" ", @mods);
+&write_file($httpd_info_cache, \%cache);
+return ($ver, \@mods, $fullver);
 }
 
 # print_line(directive, text, indent, link)
@@ -1253,7 +1385,7 @@ local $txtlen = length($txt);
 $txt = &html_escape($txt);
 print " " x $_[2];
 if ($_[3]) {
-	print "<a href=\"$_[3]\">",$txt,"</a>";
+	print &ui_link($_[3], $txt);
 	}
 else { print $txt; }
 print " " x (90 - $txtlen - $_[2] - length($lstr));
@@ -1332,7 +1464,8 @@ sub lock_apache_files
 {
 local $conf = &get_config();
 local $f;
-foreach $f (&unique(map { $_->{'file'} } @$conf)) {
+@main::locked_apache_files = &unique(map { $_->{'file'} } @$conf);
+foreach $f (@main::locked_apache_files) {
 	&lock_file($f);
 	}
 }
@@ -1341,24 +1474,27 @@ sub unlock_apache_files
 {
 local $conf = &get_config();
 local $f;
-foreach $f (&unique(map { $_->{'file'} } @$conf)) {
+foreach $f (@main::locked_apache_files) {
 	&unlock_file($f);
 	}
+@main::locked_apache_files = ( );
 }
 
 # directive_lines(directive, ...)
+# Convery a list of Apache directives into a list of lines
 sub directive_lines
 {
-local @rv;
-foreach $d (@_) {
+my @rv;
+foreach my $d (@_) {
 	next if ($d->{'name'} eq 'dummy');
+	my $indent = (" " x $d->{'indent'});
 	if ($d->{'type'}) {
-		push(@rv, "<$d->{'name'} $d->{'value'}>");
+		push(@rv, $indent."<$d->{'name'} $d->{'value'}>");
 		push(@rv, &directive_lines(@{$d->{'members'}}));
-		push(@rv, "</$d->{'name'}>");
+		push(@rv, $indent."</$d->{'name'}>");
 		}
 	else {
-		push(@rv, "$d->{'name'} $d->{'value'}");
+		push(@rv, $indent."$d->{'name'} $d->{'value'}");
 		}
 	}
 return @rv;
@@ -1389,7 +1525,7 @@ if ($httpd_modules{'core'} >= 1.301) {
 			}
 		}
 	local $out = &backquote_command("$cmd 2>&1");
-	if ($out && $out !~ /syntax\s+ok/i) {
+	if ($out && $out !~ /(syntax|Checking).*\s+ok/i) {
 		return $out;
 		}
 	}
@@ -1539,6 +1675,7 @@ else {
 	&kill_logged('HUP', $1) || return &text('restart_esig', $1);
 	&wait_for_graceful();
 	}
+&restart_last_restart_time();
 return undef;
 }
 
@@ -1553,7 +1690,7 @@ local @st = stat($errorlog);
 my $start = time();
 while(time() - $start < $timeout) {
 	sleep(1);
-	open(ERRORLOG, $errorlog);
+	open(ERRORLOG, "<".$errorlog);
 	seek(ERRORLOG, $st[7], 0);
 	local $/ = undef;
 	local $rest = <ERRORLOG>;
@@ -1588,7 +1725,7 @@ elsif (-x $config{'apachectl_path'}) {
 else {
 	# kill the process
 	$pidfile = &get_pid_file();
-	open(PID, $pidfile) || return &text('stop_epid', $pidfile);
+	open(PID, "<".$pidfile) || return &text('stop_epid', $pidfile);
 	<PID> =~ /(\d+)/ || return &text('stop_epid2', $pidfile);
 	close(PID);
 	&kill_logged('TERM', $1) || return &text('stop_esig', $1);
@@ -1667,10 +1804,11 @@ if (!&is_apache_running()) {
 		}
 	else {
 		$errorlog = &server_root($errorlog, $conf);
-		$out = `tail -5 $errorlog`;
+		$out = &backquote_command("tail -5 ".quotemeta($errorlog));
 		return "$text{'start_eafter'} : <pre>$out</pre>";
 		}
 	}
+&restart_last_restart_time();
 return undef;
 }
 
@@ -1712,39 +1850,51 @@ for(my $i=0; $i<$secs; $i++) {
 return 0;
 }
 
-# configurable_modules()
+# configurable_modules([&all-mods])
 # Returns a list of Apaches that are compiled in or dynamically loaded, and
 # supported by Webmin.
 sub configurable_modules
 {
-local ($ver, $mods) = &httpd_info(&find_httpd());
-local @rv;
-local $m;
+my ($allmods) = @_;
+$allmods ||= [ &available_modules() ];
+return grep { -r "$module_root_directory/$_.pl" } @$allmods;
+}
 
-# add compiled-in modules
-foreach $m (@$mods) {
-	if (-r "$module_root_directory/$m.pl") {
-		push(@rv, $m);
-		}
-	}
+# available_modules()
+# Returns a list of Apaches that are compiled in or dynamically loaded
+sub available_modules
+{
+my ($ver, $mods) = &httpd_info(&find_httpd());
+my @rv;
 
-# add dynamically loaded modules
-local $conf = &get_config();
-foreach $l (&find_directive_struct("LoadModule", $conf)) {
-	if ($l->{'words'}->[1] =~ /(mod_\S+)\.(so|dll)/ &&
-	    -r "$module_root_directory/$1.pl") {
+# Add compiled-in modules
+push(@rv, @$mods);
+
+# Add dynamically loaded modules
+my $conf = &get_config();
+foreach my $l (&find_directive_struct("LoadModule", $conf)) {
+	if ($l->{'words'}->[1] =~ /(mod_\S+)\.(so|dll)/) {
 		push(@rv, $1);
 		}
-	elsif ($l->{'words'}->[1] =~ /libssl\.so/ &&
-	       -r "$module_root_directory/mod_apachessl.pl") {
+	elsif ($l->{'words'}->[1] =~ /libssl\.so/) {
 		push(@rv, "mod_apachessl");
 		}
-	elsif ($l->{'words'}->[1] =~ /lib([^\/\s]+)\.(so|dll)/ &&
-	       -r "$module_root_directory/mod_$1.pl") {
+	elsif ($l->{'words'}->[1] =~ /lib([^\/\s]+)\.(so|dll)/) {
 		push(@rv, "mod_$1");
 		}
 	}
-undef(@get_config_cache);	# Cache is no longer valid
+
+# Add dynamically loaded modules
+if ($config{'apachectl_path'}) {
+	&open_execute_command(APACHE,
+		"$config{'apachectl_path'} -M 2>/dev/null", 1);
+	while(<APACHE>) {
+		if (/(\S+)_module/) {
+			push(@rv, "mod_${1}");
+			}
+		}
+	close(APACHE);
+	}
 
 return &unique(@rv);
 }
@@ -1768,7 +1918,7 @@ if ($config{'defines_file'}) {
 	# or regular name=value format
 	local %def;
 	&read_env_file($config{'defines_file'}, \%def);
-	if ($config{'defines_name'}) {
+	if ($config{'defines_name'} && $def{$config{'defines_name'}}) {
 		# Looking for var like OPTIONS='-Dfoo -Dbar'
 		local $var = $def{$config{'defines_name'}};
 		foreach my $v (split(/\s+/, $var)) {
@@ -1834,7 +1984,8 @@ if ($config{'link_dir'}) {
 	opendir(LINKDIR, $config{'link_dir'});
 	foreach my $f (readdir(LINKDIR)) {
 		if ($f ne "." && $f ne ".." &&
-		    (&resolve_links($config{'link_dir'}."/".$f) eq $file ||
+		    (&simplify_path(
+		       &resolve_links($config{'link_dir'}."/".$f)) eq $file ||
 		     $short eq $f)) {
 			&unlink_logged($config{'link_dir'}."/".$f);
 			}
@@ -1968,6 +2119,7 @@ else {
 sub is_virtualmin_domain
 {
 local ($virt) = @_;
+return 0 if ($config{'allow_virtualmin'});
 local $n = &find_directive("ServerName", $virt->{'members'});
 return undef if (!$n);
 return undef if (!&foreign_check("virtual-server"));
@@ -1979,5 +2131,169 @@ local $d = &virtual_server::get_domain_by("dom", $n);
 return $d;
 }
 
-1;
+# update_last_config_change()
+# Updates the flag file indicating when the config was changed
+sub update_last_config_change
+{
+&open_tempfile(FLAG, ">$last_config_change_flag", 0, 1);
+&close_tempfile(FLAG);
+}
 
+# restart_last_restart_time()
+# Updates the flag file indicating when the config was changed
+sub restart_last_restart_time
+{
+&open_tempfile(FLAG, ">$last_restart_time_flag", 0, 1);
+&close_tempfile(FLAG);
+}
+
+# needs_config_restart()
+# Returns 1 if a restart is needed for sure after a config change
+sub needs_config_restart
+{
+my @cst = stat($last_config_change_flag);
+my @rst = stat($last_restart_time_flag);
+if (@cst && @rst && $cst[9] > $rst[9]) {
+	return 1;
+	}
+return 0;
+}
+
+# format_config(conf-lines-ref, [indent])
+# Formats Apache config lines with default
+# four spaces of indent for each block
+sub format_config
+{
+my ($conf_lref, $indent) = @_;
+
+# Prevent formatting if not allowed in config
+return if (!&format_config_allowed());
+
+# Default single indent equals 4 spaces
+$indent ||= 4;
+$indent = " " x $indent;
+
+# At first check if Apache blocks are ballanced
+my $conf_block_opening;
+my $conf_block_closing;
+foreach my $l (@{$conf_lref}) {
+
+    # If line doesn't start with # disregard of trailing spaces
+    if ($l !~ /^\s*#/) {
+
+        # This is a new block, count it
+        if ($l =~ /(<[a-zA-Z]+).*>/) {
+            $conf_block_opening++;
+            }
+
+        # This is a new closing block, count it
+        if ($l =~ /(<\/[a-zA-Z]+).*>/) {
+            $conf_block_closing++;
+            }
+        }
+    }
+
+# If the number of closing and opening blocks
+# is the same then generate proper indents
+if ($conf_block_opening == $conf_block_closing) {
+
+    my $conf_lvl = 0;
+    foreach my $l (@{$conf_lref}) {
+        my $indent_current = $indent x $conf_lvl;
+
+        # If line doesn't start with # disregard of trailing spaces
+        if ($l !~ /^\s*#/) {
+
+            # Indent up next line if a new block
+            if ($l =~ /(<[a-zA-Z]+).*>/) {
+                $conf_lvl++;
+                }
+
+            # Indent down next line if a closing block
+            if ($l =~ /(<\/[a-zA-Z]+).*>/) {
+                $conf_lvl--;
+
+                # Change current indent right now as it is a closing block
+                $indent_current = $indent x $conf_lvl;
+                }
+            }
+
+        # Replace beginning spaces with needed indent
+        $l =~ s/^\s*/$indent_current/
+            if($l);
+        }
+    }
+}
+
+# format_config(filename, [indent])
+# Formats Apache config given file
+sub format_config_file
+{
+my ($file, $indent) = @_;
+
+# Prevent formatting if not allowed in config
+return if (!&format_config_allowed());
+
+# If file was deleted prevent recreating an empty file
+return if (!-r $file);
+
+# Lock file
+&lock_file($file);
+
+# Open file
+my $conf_lref = &read_file_lines($file);
+
+# Format
+&format_config($conf_lref, $indent);
+
+# Write file
+&flush_file_lines($file);
+
+# Unlock file
+&unlock_file($file);
+}
+
+# format_modifed_configs([test-config])
+# Formats all modifed Apache configs during the call
+sub format_modifed_config_files
+{
+my ($force_config_test) = @_;
+if($saved_conf_files) {
+	if (&format_config_allowed()) {
+		# Test config first if not already
+		# tested and don't format on error
+		if ($force_config_test) {
+			if ($config{'test_manual'} ||
+			    $config{'test_always'}) {
+				my $conf_err = &test_config();
+				if ($conf_err) {
+					return;
+					}
+				}
+			}
+		# Format indents for each file individually
+		foreach my $saved_conf_file (&unique(@{$saved_conf_files})) {
+				&format_config_file($saved_conf_file, $config{'format_config_indent'});
+			}
+		}
+	}
+}
+
+# format_config_allowed()
+# Checks if formatting config is allowed
+sub format_config_allowed
+{
+return $config{'format_config'};
+}
+
+# clear_apache_modules_cache()
+# If new Apache modules were enabled, force re-gen of the site file
+# that contains the modules cache
+sub clear_apache_modules_cache
+{
+&unlink_file($site_file);
+&create_site_file();
+&read_site_file();
+}
+
+1;
