@@ -25,6 +25,14 @@ while(<PKGINFO>) {
 			$packages{$i,'epoch'} = $1;
 			$packages{$i,'version'} = $2;
 			}
+		if ($packages{$i,'name'} =~ /^(\S+):(\S+)$/) {
+			$packages{$i,'name'} = $1;
+			$packages{$i,'arch'} = $2;
+			}
+		if ($packages{$i,'desc'} =~ /^(all|amd64|x86|arm)\s+(\S.*)/) {
+			$packages{$i,'arch'} = $1;
+			$packages{$i,'desc'} = $2;
+			}
 		$i++;
 		}
 	}
@@ -41,12 +49,13 @@ return lc($_[0]) =~ /^[a-e]/ ? "A-E" :
        lc($_[0]) =~ /^[u-z]/ ? "U-Z" : "Other";
 }
 
-# package_info(package)
+# package_info(package, [version])
 # Returns an array of package information in the order
 #  name, class, description, arch, version, vendor, installtime
 sub package_info
 {
-local $qm = quotemeta($_[0]);
+local ($pkg, $ver) = @_;
+local $qm = quotemeta($pkg);
 
 # First check if it is really installed, and not just known to the package
 # system in some way
@@ -57,10 +66,81 @@ if ($lines[$#lines] !~ /^.[ih]/) {
 	}
 
 # Get full status
-local $out = &backquote_command("dpkg --print-avail $qm 2>&1", 1);
+local $out;
+if (&has_command("apt-cache")) {
+	$qm .= "=*".quotemeta($ver) if ($ver);
+	$out = &backquote_command("apt-cache show $qm 2>&1", 1);
+	$out =~ s/[\0-\177]*\r?\n\r?\n(Package:)/\\1/;	# remove available ver
+	}
+else {
+	$out = &backquote_command("dpkg --print-avail $qm 2>&1", 1);
+	}
 return () if ($? || $out =~ /Package .* is not available/i);
+
+# Set name and class
 local @rv = ( $_[0], &alphabet_name($_[0]) );
-push(@rv, $out =~ /Description:\s+([\0-\177]*\S)/i ? $1
+
+# Get description
+push(@rv, $out =~ /Description(-en)?:\s+((.*\n)(\s+.*\n)*)/i
+	? $2
+	: $text{'debian_unknown'});
+
+# Get architecture
+push(@rv, $out =~ /Architecture:\s+(\S+)/i
+	? $1
+	: $text{'debian_unknown'});
+my $arch = $1;
+
+# Get version
+push(@rv, $out =~ /Version:\s+(\S+)/i
+	? $1
+	: $text{'debian_unknown'});
+
+# Get maintainer
+push(@rv, $out =~ /Maintainer:\s+(.*)/i
+	? &html_escape($1)
+	: $text{'debian_unknown'});
+
+# Get install time from the package info file
+my ($path) = grep { -e $_ } $arch
+	? ("/var/lib/dpkg/info/$pkg:$arch.list", "/var/lib/dpkg/info/$pkg.list")
+	: ("/var/lib/dpkg/info/$pkg.list");   # prefer multi-arch
+my $inst_epoch = $path ? (stat($path))[9] : undef;
+push(@rv, $inst_epoch ? &make_date($inst_epoch) : undef);
+
+# Get homepage
+push(@rv, $out =~ /Homepage:\s+(.*)/i ? $1 : undef);
+
+# Return info
+return @rv;
+}
+
+# virtual_package_info(package)
+# Returns an array of package information for a virtual package, usually called
+# if "package_info" returns nothing.
+sub virtual_package_info
+{
+my ($pkg) = @_;
+my $qpkg = quotemeta($pkg);
+my $apt_cache_cmd = &has_command("apt-cache");
+return ( ) if (!$apt_cache_cmd);
+my $out = &backquote_command("$apt_cache_cmd showpkg $qpkg 2>&1", 1);
+
+# Get the package that provides this virtual package
+my ($vpkg);
+if ($out =~ /Reverse Provides:\s*(\S+\s+\S+)/) {
+	my ($rpkg, $rver) = split(/\s+/, $1, 2);
+	$vpkg = $rpkg;
+	$qvpkg = quotemeta($vpkg)."=".quotemeta($rver);
+	}
+return ( ) if (!$vpkg);
+
+# Get full status
+$out = &backquote_command("$apt_cache_cmd show $qvpkg 2>&1", 1);
+$out =~ s/[\0-\177]*\r?\n\r?\n(Package:)/\\1/;	# remove available ver
+return () if ($? || $out =~ /Package .* is not available/i);
+local @rv = ( $vpkg, &alphabet_name($vpkg) );
+push(@rv, $out =~ /Description(-en)?:\s+((.*\n)(\s+.*\n)*)/i ? $2
 						   : $text{'debian_unknown'});
 push(@rv, $out =~ /Architecture:\s+(\S+)/i ? $1 : $text{'debian_unknown'});
 push(@rv, $out =~ /Version:\s+(\S+)/i ? $1 : $text{'debian_unknown'});
@@ -120,7 +200,7 @@ sub installed_file
 {
 local $qm = quotemeta($_[0]);
 local $out = &backquote_command("dpkg --search $qm 2>&1", 1);
-return 0 if ($out =~ /not found/i);
+return 0 if ($out =~ /not\s+found|no\s+path\s+found/i);
 $out =~ s/:\s+\S+\n$//;
 local @pkgin = split(/[\s,]+/, $out);
 local $real = &translate_filename($_[0]);
@@ -187,6 +267,7 @@ local $args = ($in->{'depends'} ? " --force-depends" : "").
 	      ($in->{'overwrite'} ? " --force-overwrite" : "").
 	      ($in->{'downgrade'} ? " --force-downgrade" : "");
 local $qm = quotemeta($_[0]);
+$ENV{'UCF_FORCE_CONFFOLD'} = 'YES';
 $ENV{'DEBIAN_FRONTEND'} = 'noninteractive';
 local $out = &backquote_logged("dpkg --install $args $qm 2>&1 </dev/null");
 if ($?) {
@@ -195,19 +276,71 @@ if ($?) {
 return undef;
 }
 
-# delete_package(package)
+# delete_options(package)
+# Outputs HTML for package uninstall options
+sub delete_options
+{
+print "<b>$text{'delete_purge'}</b>\n";
+print &ui_yesno_radio("purge", 0),"<br>\n";
+
+if ($update_system eq "apt") {
+	print "<b>$text{'delete_depstoo'}</b>\n";
+	print &ui_yesno_radio("depstoo", 0),"<br>\n";
+	}
+}
+
+# delete_package(package, [&options], version)
 # Totally remove some package
 sub delete_package
 {
 local $qm = quotemeta($_[0]);
 $ENV{'DEBIAN_FRONTEND'} = 'noninteractive';
-local $out = &backquote_logged("dpkg --remove $qm 2>&1 </dev/null");
+local $out;
+if ($_[1]->{'depstoo'}) {
+	# Use apt-get
+	local $flag = $_[1]->{'purge'} ? "--purge" : "";
+	$out = &backquote_logged("apt-get -y autoremove $flag $qm 2>&1 </dev/null");
+	}
+else {
+	# Use dpkg command
+	local $flag = $_[1]->{'purge'} ? "--purge" : "--remove";
+	$out = &backquote_logged("dpkg $flag $qm 2>&1 </dev/null");
+	}
 if ($? || $out =~ /which isn.t installed/i) {
 	return "<pre>$out</pre>";
 	}
 return undef;
 }
 
+# package_dependencies(name, [version])
+# Returns a list of packages that this one depends on, as hash refs
+sub package_dependencies
+{
+my ($pkg, $ver) = @_;
+my $out = &backquote_command(
+	"apt-cache show ".quotemeta($pkg)." 2>/dev/null");
+return () if ($?);
+$out =~ /\nDepends:\s+(.*)/ || return ();
+my @rv;
+foreach my $w (split(/,\s+/, $1)) {
+	if ($w =~ /^(\S+)\s+\(([<>=]+)\s+(\S+)\)/) {
+		# Depends on a specific version
+		push(@rv, { 'package' => $1,
+			    'compare' => $2,
+			    'version' => $3 });
+		}
+	elsif ($w =~ /^(\S+)$/) {
+		# Depends on some package
+		push(@rv, { 'package' => $1 });
+		}
+	elsif ($w =~ /^(\S+)\s+\|/) {
+		# Depends on one of several packages, but just show the first
+		push(@rv, { 'package' => $1 });
+		}
+	}
+return @rv;
+}
+
 sub package_system
 {
 return $text{'debian_manager'};
